/**
 * Dependencies
 */
var _ = require('lodash');
var NOOT = require('../../../../../index')('object', 'http', 'errors');
var async = require('async');
var flatten = require('flat');

var Operators = require('./../../operators/index');
var Authable = require('./../../mixins/authable');
var Queryable = require('./../../mixins/queryable');
var DefaultRoutes = require('./../../default-routes/index');
var Route = require('./../../route');
var Field = require('./../../fields/lib/field');

function removeWildCard(path) {
  return Field.removeWildcardsFromPath(path);
}
function removeFirstSegment(key) {
  return key.split('.').slice(1).join('.');
}

function getFirstSegment(key) {
  return _.first(key.split('.'));
}

function deepMergeValues(arr1, arr2) {
  if (!arr2) return arr1;
  return arr1.map(function(item, index) {
    if (_.isArray(item)) return deepMergeValues(item, arr2[index]);
    return _.merge(item, arr2[index]);
  });
}

function isFieldObjectAndPropArray(field, prop) {
 return !_.isArray(field) && typeof field === 'object' && _.isArray(field[prop]);
}

function getNestedArrayByPath(item, key) {
  if (!key) return item;
  if (_.isArray(item)) {
    return item.map(function(field) {
      var newKey = removeFirstSegment(key);
      var property = getFirstSegment(key);
      // _.isPlainObject will be false even though its an object, but arrays are _.isObjects as well
      if (isFieldObjectAndPropArray(field, property)) {
        field[property] = getNestedArrayByPath(field[property], newKey);
        return field;
      }

      return getNestedArrayByPath(field, newKey);
    });
  }

  return _.pick(item, key);
}

var filterSelectedFields = _.curry(function (keys, item) {
  var newItem = {};

  _.forEach(keys, function(key) {
    var property = getFirstSegment(key);
    var isNestedArray = _.contains(key, '.') && _.isArray(item[property]);

    if (isNestedArray) {
      var existingValue = NOOT.tryGet(newItem, property);
      var value = getNestedArrayByPath(item[property], removeFirstSegment(key));
      deepMergeValues(value, existingValue);
      NOOT.trySet(newItem, property, value);
    } else {
      NOOT.trySet(newItem, key, NOOT.tryGet(item, key));
    }
  });

  return newItem;
});

/***********************************************************************************************************************
 * @class Resource
 * @namespace NOOT.API
 * @constructor
 * @extends NOOT.Object
 * @uses NOOT.API.Queryable
 * @uses NOOT.API.Authable
 **********************************************************************************************************************/
var Resource = NOOT.Object.extend(Authable).extend(Queryable).extend({

  /**
   * @property methods
   * @type Array
   */
  methods: null,

  /**
   * @property detailMethods
   * @type Array
   */
  detailMethods: null,

  /**
   * @property manyMethods
   * @type Array
   */
  manyMethods: null,

  /**
   * @property defaultStatusCode
   * @type Number
   */
  defaultStatusCode: null,

  /**
   * @property allowedResponseTypes
   * @type Array
   */
  allowedResponseTypes: null,

  /**
   * @property defaultResponseType
   * @type String
   */
  defaultResponseType: null,

  /**
   * @property fields
   * @type Object
   */
  fields: null,

  /**
   * @property fieldsPath
   * @type Array
   * @readOnly
   */
  get fieldsPaths() { return NOOT.isObject(this.fields) ? Object.keys(this.fields) : []; },

  /**
   * Constructor
   */
  init: function() {
    NOOT.defaults(this, Resource.DEFAULTS);
    NOOT.required(this, 'api');
    if (!(this.api instanceof require('./../../api'))) throw new Error('Not a NOOT.API');
    this.fields = this.fields || this.getFields() || {};
    this.computeQueryable();
    this._buildMethods();
    this._buildRoutes();
  },

  /**
   * Will be called b the constructor in case `fields` is not defined. This method allows you to process
   *
   * @method getFields
   */
  getFields: function() {},

  /**
   * In charge of parsing and validating default routes methods defined in `methods`, `detailMethods` and `manyMethods`.
   *
   * @method _buildMethods
   * @private
   */
  _buildMethods: function() {
    var detail = this.detailMethods;
    var many = this.manyMethods;
    var all = this.methods;

    if (detail) {
      detail.forEach(Resource.validateDefaultRouteMethod.bind(Resource));
      detail = detail.map(function(method) { return method.toLowerCase(); });
    }

    if (many) {
      many.forEach(Resource.validateDefaultRouteMethod.bind(Resource));
      many = many.map(function(method) { return method.toLowerCase(); });
    }

    if (all) {
      all.forEach(Resource.validateDefaultRouteMethod.bind(Resource));
      all = all.map(function(method) { return method.toLowerCase(); });
    } else {
      all = this.constructor.DEFAULT_ROUTES_METHODS;
    }

    // Assign final values and make them readOnly
    NOOT.makeReadOnly(this, 'detailMethods', detail || all);
    NOOT.makeReadOnly(this, 'manyMethods', many || all);
    NOOT.makeReadOnly(this, 'methods', all);
  },

  /**
   * In charge of creating the final `routes` property. This method will create an array of NOOT.API.Route instances
   * to be used by the resource. The final array will contain both default and user defined routes.
   *
   * Route classes need to be instantiates twice, because this is the only way to compare them.
   *
   * @method _buildRoutes
   * @private
   */
  _buildRoutes: function() {
    var self = this;
    var routes = this.routes || [];

    var controllRoutes = routes.map(function(routeClass) {
      return routeClass.create({ resource: self, __queryableParent: self });
    });

    var isRouteAlreadyDeclared = function(routeClass) {
      var instance = routeClass.create({ resource: self, __queryableParent: self });
      return !!_.find(controllRoutes, { method: instance.method, path: instance.path });
    };

    this.detailMethods.forEach(function(method) {
      if (DefaultRoutes[method].prototype.isDetailable) {
        var routeClass = DefaultRoutes[method].extend({ isDetail: true });

        if (!isRouteAlreadyDeclared(routeClass)) routes.push(routeClass);
      }
    });

    this.manyMethods.forEach(function(method) {
      var routeClass = DefaultRoutes[method].extend({ isDetail: false });

      if (!isRouteAlreadyDeclared(routeClass)) routes.push(routeClass);
    });

    this.routes = routes.map(function(routeClass) {
      if (!Route.detect(routeClass)) throw new Error('Not an NOOT.API.Route');
      return routeClass.create({ resource: self, __queryableParent: self });
    });
  },

  /**
   * Cleaning `data` by `_id` and `select` fields
   *
   * @method _cleanupBySelection
   * @param {Object} data
   * @param {Object} select
   * @returns {Array}
   * @private
   */
  _cleanupBySelection: function(data, select) {
    var keys = _.chain(['_id']).concat(_.keys(select)).uniq().value();

    if (_.isArray(data)) {
      return _.map(data, filterSelectedFields(keys));
    }

    if (_.isObject(data)) {
      return filterSelectedFields(keys, data);
    }

    return data;
  },

  /**
   * In charge of formatting the response that will be sent. Here you can format the stack's package and define how
   * the response will look like.
   *
   * @method formatResponsePackage
   * @param {NOOT.API.Stack} stack
   * @param {Function} [callback]
   */
  formatResponsePackage: function(stack, callback) {
    callback = callback || stack.next;
    var pack = stack.package;
    if (!pack.messages.length) delete pack.messages;
    // if there is a select qs, we override `minimize` settings
    if (NOOT.tryGet(stack, 'query.raw.select')) {
      pack.data = this._cleanupBySelection(pack.data, stack.query.select);
    }
    stack.package = _.pick(pack, ['data', 'error', 'message', 'messages', 'meta']);
    return callback();
  },

  /**
   * In charge of calling the right handler in order to respond to the request. If no `accept` header is provided by
   * the request, the `defaultResponseType` is used. If the provided `accept` is not supported, a
   * NOOT.Errors.NotImplemented error is sent as a response.
   *
   * @method sendResponse
   * @param {NOOT.API.Stack} stack
   */
  sendResponse: function(stack) {
    stack.res.status(stack.statusCode || this.defaultStatusCode);

    var type = stack.req.accepts(this.allowedResponseTypes) || this.defaultResponseType;

    switch (type) {
      case 'json': return this.sendJSON(stack);
      default:
        stack.pushMessage(this.api.messagesProvider.unsupportedResponseType(type));
        return stack.next(new NOOT.Errors.NotImplemented());
    }
  },

  /**
   * Sends stack's package as a JSON response.
   *
   * @method sendJSON
   * @param {NOOT.API.Stack} stack
   */
  sendJSON: function(stack) {
    return stack.res.json(stack.package);
  },

  /**
   * Middleware to parse and validate stack's filter. If `query.filter` contains a field that is not `filterable`, a
   * NOOT.Errors.Forbidden error is passed to the callback.
   *
   * @method parseQueryFilter
   * @param {NOOT.API.Stack} stack
   * @param {Function} [callback]
   */
  parseQueryFilter: function(stack, callback) {
    var self = this;
    var ret = {};
    var map = {};
    var fields = this.fields;
    var filter = stack.query.filter;
    callback = callback || stack.next;

    return async.each(Object.keys(filter), function(filterName, cb) {
      var split = filterName.split(self.constructor.OPERATOR_SEPARATOR);
      var publicPath = split[0];
      var operatorName = split[1] || self.constructor.EQUALITY_OPERATOR;

      var operator = Operators[operatorName];
      var field = _.find(fields, function(field) {
        if (field.isAny) return (new RegExp('^' + NOOT.toRegExpString(field.publicPath))).test(publicPath);
        return publicPath === field.publicPath;
      });

      if (!field) {
        stack.pushMessage(self.api.messagesProvider.forbiddenField(publicPath));
        return callback(new NOOT.Errors.Forbidden());
      }

      if (!field.validateOperator(operatorName)) {
        stack.pushMessage(self.api.messagesProvider.forbiddenOperator(publicPath, operatorName));
        return callback(new NOOT.Errors.Forbidden());
      }

      if (!operator) {
        stack.pushMessage(self.api.messagesProvider.unsupportedOperator(publicPath, operatorName));
        return callback(new NOOT.Errors.BadRequest());
      }

      map[publicPath] = map[publicPath] || {};
      operator.parseFromQueryString(filter[filterName], field.parseFromQueryString.bind(field), function(err, value) {
        if (err) return cb(err);

        map[publicPath][operatorName] = value;
        return cb();
      });

    }, function(err) {
      if (err) return callback(err);

      for (var path in map) {
        var pathFilter = map[path];
        if (pathFilter.hasOwnProperty(self.constructor.EQUALITY_OPERATOR)) ret[path] = pathFilter.eq;
        else ret[path] = pathFilter;
      }

      stack.query.filter = ret;

      return callback();
    });
  },

  /**
   * Middleware to parse and validate stack's select. If `query.select` contains a field that is not `selectable`, a
   * NOOT.Errors.Forbidden error is passed to the callback.
   *
   * @method parseQuerySelect
   * @param {NOOT.API.Stack} stack
   * @param {Function} [callback]
   */
  parseQuerySelect: function(stack, callback) {
    var self = this;
    callback = callback || stack.next;
    var select = stack.query.select.map(removeWildCard);
    var selectable = stack.selectable.map(removeWildCard);
    var isValid = true;

    if (select && select.length) {
      var final = [];
      select.forEach(function(fieldName) {
        var shouldExclude =  fieldName[0] === self.constructor.EXCLUSION_CHARACTER;
        var rawFieldName = fieldName.replace(new RegExp('^' + self.constructor.EXCLUSION_CHARACTER), '');
        var childs = selectable.filter(function(selectableName) {
          return selectableName.split('.')[0] === rawFieldName;
        }).map(function(selectableName) {
          return shouldExclude ? self.constructor.EXCLUSION_CHARACTER + selectableName : selectableName;
        });
        if (childs.length) {
          final = final.concat(childs);
        } else {
          if (_.contains(selectable, rawFieldName)) {
            final.push(Field.removeWildcardsFromPath(fieldName));
          } else {
            isValid = false;
            stack.pushMessage(self.api.messagesProvider.forbiddenField(rawFieldName, 'select'));
          }
        }
      });
      stack.query.select = final;
    }

    return callback(isValid ? null : new NOOT.Errors.Forbidden());
  },

  /**
   * Middleware to parse and validate stack's sort. If `query.sort` contains a field that is not `sortable`, a
   * NOOT.Errors.Forbidden error is passed to the callback.
   *
   * @method parseQuerySort
   * @param {NOOT.API.Stack} stack
   * @param {Function} [callback]
   */
  parseQuerySort: function(stack, callback) {
    var self = this;
    callback = callback || stack.next;
    var sort = stack.query.sort;
    var sortable = stack.sortable;
    var isValid = true;

    if (sort && sort.length) {
      sort.forEach(function(fieldName) {
        fieldName = fieldName.replace(new RegExp('^' + self.constructor.EXCLUSION_CHARACTER), '');
        if (!_.contains(sortable, fieldName)) {
          isValid = false;
          stack.pushMessage(self.api.messagesProvider.forbiddenField(fieldName, 'sort'));
        }
      });
    }

    return callback(isValid ? null : new NOOT.Errors.Forbidden());
  },

  /**
   * Middleware to parse and validate stack's body. If `body` contains a field that is not `writable`, a
   * NOOT.Errors.Forbidden error is passed to the callback.
   *
   * @method parseQueryBody
   * @param {NOOT.API.Stack} stack
   * @param {Function} [callback]
   */
  parseQueryBody: function(stack, callback) {
    var self = this;
    var writable = stack.writable;
    callback = callback || stack.next;
    var isValid = true;
    var body = NOOT.isArray(stack.body) ? stack.body : [stack.body];

    body.forEach(function(item) {
      Object.keys(flatten(item)).forEach(function(key) {
        var wildcarded = Field.replaceReferenceWithWildcard(key);
        var unaddressed = Field.replaceReferenceWithWildcard(key).replace(/\.\d+$/, '');
        var isValidField = _.contains(writable, key) ||
          _.contains(writable, wildcarded) ||
          _.contains(writable, unaddressed) ||
          !!_.find(self.fields, function (field) {
            if (field.isAny) {
              return (new RegExp('^' + NOOT.toRegExpString(field.path))).test(unaddressed);
            }
          });

        if (!isValidField) {
          isValid = false;
          stack.pushMessage(self.api.messagesProvider.forbiddenField(key, 'write'));
        }
      });
    });

    return callback(isValid ? null : new NOOT.Errors.Forbidden());
  }

}, {

  /**
   * @property DEFAULTS
   * @type Object
   * @static
   */
  DEFAULTS: {
    get methods() { return DefaultRoutes.supportedMethods; },
    get defaultStatusCode() { return NOOT.HTTP.OK; },
    get allowedResponseTypes() { return ['json']; },
    get defaultResponseType() { return 'json'; }
  },

  /**
   * @property OPERATOR_SEPARATOR
   * @type String
   * @static
   */
  OPERATOR_SEPARATOR: '__',

  /**
   * @property EQUALITY_OPERATOR
   * @type String
   * @static
   */
  EQUALITY_OPERATOR: 'eq',

  /**
   * @property EXCLUSION_CHARACTER
   * @type String
   * @static
   */
  EXCLUSION_CHARACTER: '-',

  /**
   * Validates an HTTP method against those that are supported by NOOT.API.DefaultRoutes. If the parameter method is not
   * supported, an error will be thrown.
   *
   * @method validateDefaultRouteMethod
   * @param {String} method
   * @static
   */
  validateDefaultRouteMethod: function(method) {
    if (!_.contains(DefaultRoutes.supportedMethods, method.toLowerCase())) {
      throw new Error('Invalid method for auto generated route: ' + method);
    }
  }

});


/**
 * @exports
 */
module.exports = Resource;
